/*
 * nevm
 *
 * Minimalistic ZMODEM Implementation for Download
 * of Screenshots and Measurements over RS232
 */

#include <stdio.h>
#include <string.h>

#include "hardware_t.h"
#include "tc_vars.h"
#include "flash_t.h"
#include "display_t.h"
#include "zmodem.h"

/*
 * (zm_)crctab calculated by Mark G. Mendel, Network Systems Corporation.
 * (zm_)updcrc macro derived from article Copyright (C) 1986 Stephen Satchell.
 *
 * Programmers may incorporate any or all code into their programs,
 * giving proper credit within the source. Publication of the
 * source routines is permitted so long as proper credit is given
 * to Stephen Satchell, Satchell Evaluations and Chuck Forsberg,
 * Omen Technology.
 */

#define zm_updcrc(cp, crc) ( zm_crctab[((crc >> 8) & 255)] ^ (crc << 8) ^ cp)

static const unsigned short zm_crctab[256] = {
    0x0000,  0x1021,  0x2042,  0x3063,  0x4084,  0x50a5,  0x60c6,  0x70e7,
    0x8108,  0x9129,  0xa14a,  0xb16b,  0xc18c,  0xd1ad,  0xe1ce,  0xf1ef,
    0x1231,  0x0210,  0x3273,  0x2252,  0x52b5,  0x4294,  0x72f7,  0x62d6,
    0x9339,  0x8318,  0xb37b,  0xa35a,  0xd3bd,  0xc39c,  0xf3ff,  0xe3de,
    0x2462,  0x3443,  0x0420,  0x1401,  0x64e6,  0x74c7,  0x44a4,  0x5485,
    0xa56a,  0xb54b,  0x8528,  0x9509,  0xe5ee,  0xf5cf,  0xc5ac,  0xd58d,
    0x3653,  0x2672,  0x1611,  0x0630,  0x76d7,  0x66f6,  0x5695,  0x46b4,
    0xb75b,  0xa77a,  0x9719,  0x8738,  0xf7df,  0xe7fe,  0xd79d,  0xc7bc,
    0x48c4,  0x58e5,  0x6886,  0x78a7,  0x0840,  0x1861,  0x2802,  0x3823,
    0xc9cc,  0xd9ed,  0xe98e,  0xf9af,  0x8948,  0x9969,  0xa90a,  0xb92b,
    0x5af5,  0x4ad4,  0x7ab7,  0x6a96,  0x1a71,  0x0a50,  0x3a33,  0x2a12,
    0xdbfd,  0xcbdc,  0xfbbf,  0xeb9e,  0x9b79,  0x8b58,  0xbb3b,  0xab1a,
    0x6ca6,  0x7c87,  0x4ce4,  0x5cc5,  0x2c22,  0x3c03,  0x0c60,  0x1c41,
    0xedae,  0xfd8f,  0xcdec,  0xddcd,  0xad2a,  0xbd0b,  0x8d68,  0x9d49,
    0x7e97,  0x6eb6,  0x5ed5,  0x4ef4,  0x3e13,  0x2e32,  0x1e51,  0x0e70,
    0xff9f,  0xefbe,  0xdfdd,  0xcffc,  0xbf1b,  0xaf3a,  0x9f59,  0x8f78,
    0x9188,  0x81a9,  0xb1ca,  0xa1eb,  0xd10c,  0xc12d,  0xf14e,  0xe16f,
    0x1080,  0x00a1,  0x30c2,  0x20e3,  0x5004,  0x4025,  0x7046,  0x6067,
    0x83b9,  0x9398,  0xa3fb,  0xb3da,  0xc33d,  0xd31c,  0xe37f,  0xf35e,
    0x02b1,  0x1290,  0x22f3,  0x32d2,  0x4235,  0x5214,  0x6277,  0x7256,
    0xb5ea,  0xa5cb,  0x95a8,  0x8589,  0xf56e,  0xe54f,  0xd52c,  0xc50d,
    0x34e2,  0x24c3,  0x14a0,  0x0481,  0x7466,  0x6447,  0x5424,  0x4405,
    0xa7db,  0xb7fa,  0x8799,  0x97b8,  0xe75f,  0xf77e,  0xc71d,  0xd73c,
    0x26d3,  0x36f2,  0x0691,  0x16b0,  0x6657,  0x7676,  0x4615,  0x5634,
    0xd94c,  0xc96d,  0xf90e,  0xe92f,  0x99c8,  0x89e9,  0xb98a,  0xa9ab,
    0x5844,  0x4865,  0x7806,  0x6827,  0x18c0,  0x08e1,  0x3882,  0x28a3,
    0xcb7d,  0xdb5c,  0xeb3f,  0xfb1e,  0x8bf9,  0x9bd8,  0xabbb,  0xbb9a,
    0x4a75,  0x5a54,  0x6a37,  0x7a16,  0x0af1,  0x1ad0,  0x2ab3,  0x3a92,
    0xfd2e,  0xed0f,  0xdd6c,  0xcd4d,  0xbdaa,  0xad8b,  0x9de8,  0x8dc9,
    0x7c26,  0x6c07,  0x5c64,  0x4c45,  0x3ca2,  0x2c83,  0x1ce0,  0x0cc1,
    0xef1f,  0xff3e,  0xcf5d,  0xdf7c,  0xaf9b,  0xbfba,  0x8fd9,  0x9ff8,
    0x6e17,  0x7e36,  0x4e55,  0x5e74,  0x2e93,  0x3eb2,  0x0ed1,  0x1ef0
};

static volatile char uart_buf[256];
static volatile char uart_buf_p;

static void
zm_send_hexhdr(int type, char data[4])
{
	char buf[64];
	int i;
	short crc;

	crc = zm_updcrc(type, 0);
	for (i = 0; i < 4; i++)
		crc = zm_updcrc(255 & data[i], crc);
	crc = zm_updcrc(0, zm_updcrc(0, crc));

	(void)snprintf(buf, sizeof(buf), "**\030B%.2x%.2x%.2x%.2x%.2x%.2x%.2x\r\212",
			type, data[0], data[1], data[2], data[3],
			(char)(crc >> 8), (char)crc);

	for (i = 0; i < strlen(buf); i++)
		(void)putchar(buf[i]);
}

static void
zm_send_zdle(int c)
{
	/*
	 * Chapter 7.2 ZMODEM Spec
	 *
	 * We always escape 015 and 0215 in order
	 * not having to bother with previously sent
	 * chars.
	 */

	if (	c == 030 || c == 020 || c == 0220 || c == 021
		|| c == 0221 || c == 023 || c == 0223
		|| c == 015 || c == 0215) {
		(void)putchar(0x18);
		c ^= 0x40;
	}
	(void)putchar(c);
}

static void
zm_send_binhdr(int type, char data[4])
{
	int i;
	short crc;

	(void)putchar('*');
	(void)putchar(0x18);
	(void)putchar('A');

	crc = zm_updcrc(type, 0);
	zm_send_zdle(type);

	for (i = 0; i < 4; i++) {
		crc = zm_updcrc(255 & data[i], crc);
		zm_send_zdle(data[i]);
	}

	crc = zm_updcrc(0, zm_updcrc(0, crc));
	zm_send_zdle((char)(crc >> 8));
	zm_send_zdle((char)crc);
}

static void
zm_send_data(char frameend, char *buf, int len)
{
	int i;
	short crc;

	crc = 0;
	for (i = 0; i < len; i++) {
		crc = zm_updcrc(255 & buf[i], crc);
		zm_send_zdle(buf[i]);
	}

	(void)putchar(0x18);
	crc = zm_updcrc(255 & frameend, crc);
	(void)putchar(frameend);

	crc = zm_updcrc(0, zm_updcrc(0, crc));
	zm_send_zdle((char)(crc >> 8));
	zm_send_zdle((char)crc);
}

static int
zm_get_hexhdr(void)
{
	int b[2], i;
	char c;

	for (i = 0; i < 2; i++) {
		c = (char)nr_rxchar();

		if (c >= '0' && c <= '9')
			b[i] = c - '0';
		else if (c >= 'a' && c <= 'f')
			b[i] = 10 + c - 'a';
		else if (c >= 'A' && c <= 'F')
			b[i] = 10 + c - 'A';
		else
			b[i] = 0;
	}

	return (b[1] | (b[0] << 4));
}

static int
zm_get_binhdr(void)
{
	char c;

	if ((c = (char)nr_rxchar()) == 0x18)
		if (((c = (char)nr_rxchar()) & 0x60) == 0x40)
			c &= 0xBF;

	return c;
}

static int
zm_get_header(void)
{
	int i, c, n, type;
	short crc;

	n = 0;
	while ((c = nr_rxchar()) != '*') {
		if (c == 0x18) {
			if (++n == 5)
				return -1;
		} else
			n = 0;
	}

	while ((c = nr_rxchar()) == '*')
		;

	crc = 0;
	type = -1;
	if (c == 0x18) {
		if ((c = nr_rxchar()) == 'B') {
			// hex header follows
			type = zm_get_hexhdr();
			for (i = 0; i < 6; i++)
				(void)zm_get_hexhdr();
		} else {
			// binary header follows
			type = zm_get_binhdr();
			for (i = 0; i < 6; i++)
				(void)zm_get_binhdr();
		}
	}

	return type;
}

static int
rd_char(void)
{
	int c, i;

	i = 0;
	while ((c = GETCH()) == -1 && i < 40) {
		i++;
		nr_delay(100);
	}

	if (i >= 40) {
		(void)printf("TIMEOUT!\r\n");
		return '!';
	}

	return c;
}

void
zm_test(void)
{
	char data[4], buf[1024];
	int i, r, off;
	int c;

	uart_buf_p = -1;

	(void)printf("Waiting for chars, end with \"!\"\r\n");

	while ((c = rd_char()) != '!')
		(void)printf("Read dec %d (char %c)\r\n", c, c);

	(void)printf("Waiting for string, terminated by \"!\"\r\n");

	i = 0;
	while ((c = rd_char()) != '!' && i < sizeof(buf) - 1) {
		buf[i] = c;
		i++;
	}

	buf[i] = '\0';
	(void)printf("Buffer is (string %s)\r\n", buf);

	return;

	(void)putchar('r');
	(void)putchar('z');
	(void)putchar('\r');

	// ZRQINIT ->
	(void)memset(data, 0, sizeof(data));
	zm_send_hexhdr(0, data);
	(void)putchar(0x11);

	// ZRINIT <-
	if (zm_get_header() != 1)
		goto zm_error;

	// ZFILE ->
	zm_send_binhdr(4, data);

	(void)snprintf((char *)buf, sizeof(buf), "%s%c%c", "foo3.txt", 0, 0);
	zm_send_data('k', buf, strlen((char *)buf) + 2);
	(void)putchar(0x11);

	// ZRPOS <-
	while ((r = zm_get_header()) == 1)
		;
	if (r != 9)
		goto zm_error;

	/*  We ignore position and assume 0 */
	(void)fflush(stdin);

	// ZDATA ->
	off = 0;
	for (i = 0; i < 4; i++)
		data[i] = (char)(off >> (8*i));
	zm_send_binhdr(10, data);

	(void)snprintf((char *)buf, sizeof(buf), "Erste Zeile.\n");
	zm_send_data('k', buf, strlen((char *)buf));
	off += strlen((char *)buf);

	if (zm_get_header() != 3)
		goto zm_error;

	for (i = 0; i < 4; i++)
		data[i] = (char)(off >> (8*i));
	zm_send_binhdr(10, data);

	(void)snprintf((char *)buf, sizeof(buf), "Zweite Zeile.\n");
	zm_send_data('h', buf, strlen((char *)buf));
	off += strlen((char *)buf);

	// ZEOF ->
	for (i = 0; i < 4; i++)
		data[i] = (char)(off >> (8*i));
	zm_send_binhdr(11, data);

	// ZRINIT <-
	if (zm_get_header() != 1)
		goto zm_error;

	// ZFIN ->
	(void)memset(data, 0, sizeof(data));
	zm_send_hexhdr(8, data);

	// ZFIN <-
	if (zm_get_header() != 8)
		goto zm_error;

	/* Over And Out */
	(void)putchar('O');
	(void)putchar('O');

	return;

zm_error:
	(void)printf("-- ABORT --\r\n");
}

void
zm_buf(int c)
{

	(void)printf("ZM_BUF() got called!\n");

	if (++uart_buf_p >= sizeof(uart_buf));
		uart_buf_p = sizeof(uart_buf) - 1;
	uart_buf[uart_buf_p] = c;
}

static int
GETCH(void)
{

	if (uart_buf_p < 0)
		return -1;
	return uart_buf[uart_buf_p--];
}
